/*
 * tokenize.c -- An HP-2000 Timeshared BASIC System tape manipulation program.
 *
 * Copyright (c) 2005, Scott G. Taylor
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * SCOTT G. TAYLOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Scott G. Taylor shall not
 * be used in advertising or otherwise to promote the sale, use or other dealings
 * in this Software without prior written authorization from Scott G. Taylor.
 */

#include <ctype.h>
/*#include <errno.h>*/
/*#include <fcntl.h>*/
/*#include <math.h>*/
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/param.h>
/*#include <sys/stat.h>*/
#include <sys/types.h>

#include "config.h"

/*#include "tape.h"*/
#include "tsb.h"
#include "hptsb.h"
#include "hptoken.h"
#include "decl.h"
/*#include "version.h"*/

/*
 * MAXPROGSIZE is the maximum number of words allowed for a tokenized
 * program to run in memory.  Maximum sizes per TSB version are:
 * 
 *	HP-2000A		5120
 *	HP-2000B		5120
 *	HP-2000C		10240
 *	HP-2000C'		10240
 *	HP-2000E		4180
 *	HP-2000F/F'		10240
 *	HP-2000 TSC-200/300	10240
 *	HP-2000/ACCESS		10240
 */

#define MAXPROGSIZE	10240

struct Program_Space {
	uint32_t	startloc;
	uint32_t	len;
	uint8_t		mem [MAXPROGSIZE * 2];
	uint32_t	symtabsize;
	uint8_t		symtab [MAXPROGSIZE * 2];
} ps;

/* HACKS FOR LINKING */

#define MAXENTRIES      512
 
uint32_t        last_progID = 0;

uint32_t        nentries = 0;
uint8_t         rentry[MAXENTRIES][12];

/* END HACKS */

uint8_t * get_next_character (uint8_t *);
uint8_t * eval_integer (uint8_t *, uint32_t *);
uint8_t * match_token (uint8_t *, uint32_t *);
uint32_t tokenize (uint8_t *, struct Program_Space *);

uint32_t
ltr (uint8_t *buf, uint8_t *ltr, uint8_t *ltr2)
{
}


uint8_t *
get_nonblank (uint8_t * buf)
{
	for (;*buf && *buf == ' '; buf++);
	return (buf);
}

uint8_t *
get_operand (uint8_t ** buf, uint8_t *st)
{
	uint8_t			*opti;
	struct BASIC_Operators	*bop = Bop_ACCESS;
	uint32_t		i;

	*buf = get_nonblank (*buf);
	for (i = 0; i < 077; i++) {
		opti = get_nonblank (Bop_ACCESS[i].ascii);
		if (bcmp(*buf, opti)) {
		
	
int32_t		msflag, maxcr, pflag, sflag, dflag;
int32_t		stlen;
uint32_t
tokenize (uint8_t * line, struct Program_Space *ps)
{
	uint8_t		*lp = line;
	uint8_t		*nlp;
	uint8_t		*mp;
	uint8_t		*m;
	uint8_t		*sp;
	uint8_t		*st;
	uint8_t		statement[512];
	uint8_t		varname;
	uint8_t		varnum;
	uint8_t		vartype;

	uint32_t	lineno = 0;
	uint32_t	sttype;

	if (ps != NULL) {
		m = ps->mem;
		mp = m + ps->len;
		st = ps->symtab;
		sp = st + ps ->symtabsize;
	}

	/* Determine sequence number */

	if ((nlp = eval_integer (lp, &lineno)) == NULL)
		return (1);		/* Invalid line number */
	lp = nlp;
	fprintf (stderr, "*** Line number is: %d\n", lineno);

	lp = match_token (lp, &sttype);
	if (sttype == 073)
		fprintf (stderr, "Implied LET matched.   Cool.\n");
	if (sttype == 9999)	/* Normally, implied let matches automatically */
		sttype = 073;
	if (sttype < 0)
		return (2);

	fprintf (stderr, "*** Statement Type is: (%o) %s\n",
			sttype, Bst_ACCESS[sttype].ascii);

	msflag = -9;
	maxcr = 3;
	pflag = 0;
	sflag = 0;
	dflag = 0;

	/*
	 * Statement Syntax Parsing.
	 */

	switch (sttype) {
	case 044:	/* IMAGE */
		stlen = (2 + strlen(lp) / 2) + (strlen(lp) % 2 != 0) + 1;
		statement[0] = lineno >> 8;			/* Statement No, upper byte */
		statement[1] = lineno & 0xff;			/* Statement No, lower byte */
		statement[2] = stlen  >> 8;			/* Statement len, upper byte */
		statement[3] = stlen  & 0xff;			/* Statement len, lower byte */
		statement[4] = (sttype << 9) >> 8;		/* Statement type */
		statement[5] = strlen(lp);			/* String length */
		memcpy(statement + 6, lp, strlen(lp));		/* String */
		if (strlen(lp) % 2 != 0)			/* If string spills over into */
			*(statement + 6 + strlen(lp)) = 0;	/*   next word, copy last byte */
		break;
	case 051:	/* REM */
	case 070:	/* FILES */
		/*stlen = (2 + strlen(lp) / 2) + (strlen(lp) % 2 != 0) + 1;*/
		stlen = (2 + strlen(lp) / 2) + 1;
		statement[0] = lineno >> 8;			/* Statement No, upper byte */
		statement[1] = lineno & 0xff;			/* Statement No, lower byte */
		statement[2] = stlen  >> 8;			/* Statement len, upper byte */
		statement[3] = stlen  & 0xff;			/* Statement len, lower byte */
		statement[4] = (sttype << 9) >> 8;		/* Statement type */
		memcpy(statement + 5, lp, strlen(lp));		/* String */
		if (strlen(lp) % 2 == 0)			/* If string spills over into */
			*(statement + 5 + strlen(lp)) = 0;	/*   next word, copy last byte */
		break;
	case 046:	/* LET */
	case 073:	/* Implied LET */
		lp = get_next_character (lp);
		if (!isalpha(*lp))
			return (3);	/* Not a variable, so bail. */
		varname = *lp++;
		varnum = 99;
		vartype = 0;
		/*
		 * Variable types:
		 *	0	A
		 *	1	A1
		 *	2	A[10]
		 *	3	A1[10]
		 *	5	A$
		 *	6	A1$
		 *	9	FNA
		 *	10	FNA1
		 */

		if (isdigit (*lp)) {
			varnum = *lp++ - '0';
			vartype = 1;
		}
		if (*lp == '$')
			vartype += 5;
		else if (*lp == '[' || *lp == '(')
			vartype += 2;
		if ((vartype == 1 || vartype == 6 || vartype == 10) && varnum > 2)
			return (4);

		fprintf (stderr, "Variable found: %c%c%c\n",
				varname,
				varnum == 99 ? 0 : varnum + '0',
				(vartype == 5 || vartype == 6 ) ? '$' : 0);

		stlen = 4;

		statement[0] = lineno >> 8;			/* Statement No, upper byte */
		statement[1] = lineno & 0xff;			/* Statement No, lower byte */
		statement[2] = stlen  >> 8;			/* Statement len, upper byte */
		statement[3] = stlen  & 0xff;			/* Statement len, lower byte */
		statement[4] = (sttype << 9) >> 8;		/* Statement type */
		statement[5] = varname << 4;

		break;
		
	}

	getoperand (lp);

	fprintf (stderr, "Statement length is: %d bytes\n", WTOI(statement + 2) * 2 + 2);
	Pentry.block = malloc (WTOI(statement + 2) * 2 + 2);
	memcpy(Pentry.block, statement, WTOI(statement + 2) * 2 + 2);
	Pentry.type = 0;
	Pentry.startloc = 02000;
	Pentry.filesize = WTOI(statement + 2) + 2;
	Pentry.symtabstart = 02000 + WTOI(statement + 2);
	verbose = 99;
	debug = 99;
	convert_program (&Pentry);
	return 0;
		
}

#ifdef NOTYET
uint8_t *
fsc (uint8_t *line, uint32_t sallow)
{
	uint32_t	sstak;
	uint32_t	maxcr;
	uint32_t	uflag;
	uint8_t		* lp = line;

/* #FSC */
	sstak = 0;
	maxcr = 3;
/* FSC1 */
	uflag = 0xffff;
/* FSC2 */
	
	lp = get_next_character (lp);
	if (!isalpha(*lp))
}
#endif

uint8_t *
match_token (uint8_t * line, uint32_t *opcode)
{
	uint32_t	i, j;
	uint8_t		*nlp;

	for (i = 0; i < 077; i++) {
		nlp = line;
		for (j = 0; j < strlen(Bst_ACCESS[i].ascii); j++) {
			nlp = get_next_character (nlp);
			if (*nlp != Bst_ACCESS[i].ascii[j]) {
				nlp = NULL;
				break;
			}
			nlp++;
		}
		if (nlp != NULL) {
			*opcode = i;
			return nlp;
		}
	}
	*opcode = 9999;
	return line;
}

uint8_t *
eval_integer (uint8_t * line, uint32_t *value)
{
	uint8_t		*lp = line;
	uint8_t		ndigits = 0;

	*value = 0;
	while (*lp != 0) {
		lp = get_next_character (lp);
		if (!isdigit(*lp))
			break;
		*value = *value * 10 + (*lp - '0');
		ndigits++;
		lp++;
	}
	if (ndigits)
		return (lp);
	return NULL;
}

uint8_t *
get_next_character (uint8_t *line)
{
	while (*line != 0 && (*line == ' ' || *line == '\t'))
		line++;
	return line;
}

int
main(int argc, char **argv)
{
	uint8_t	line[512];
	fprintf (stderr, "Enter a line: ");
	fgets (line, 511, stdin);
	line[strlen(line) - 1] = 0;
	fprintf (stderr, "--Line is: %s\n", line);


	tokenize (line, NULL);
	exit(0);
}
